//
// ReadGNUMathProg.cpp
//
//  Copyright 2003  Brady Hunsaker 
//  Created:  2003 May 17
//
//  License:  You may use this file under the terms of your choice of 
//    the GNU General Public License (GPL) version 2, or the 
//    Common Public License (CPL), used by the COIN-OR project.
//
//
//    Function that reads a model and data in the GNU MathProg modeling
// language into an OsiSolverInterface object, using the API from GLPK.
//
//    Note that GLPK uses indices starting at 1, while COIN-OR/OSI uses
// indices starting at 0.
//

#include "OsiSolverInterface.hpp"
#include "CoinPackedMatrix.hpp"

extern "C"
{
  #include "glpmpl.h"
}


//  readGNUMathProg
//
//  Arguments:
//    si -- pointer to a solver interface; any information will be
//            overwritten
//    modelfile -- filename with the GNU MathProg model, optionally including
//            the data section
//    datafile -- optional argument for filename of the data section
//
//  Return codes:
//    1   Model successfully loaded and is LP
//    2   Model successfully loaded and is IP/MIP
//    -1  Errors loading model
//

int readGNUMathProg ( OsiSolverInterface *si, char * modelfile, 
		      char * datafile = NULL)
{
  // initialize GNU MathProg parser
  MPL *model = mpl_initialize();

  // try to read the model; the data may also be in this file
  int errCode = mpl_read_model(model, modelfile);

  if (errCode < 1 || errCode > 2)  // errors reading the model
    {
      mpl_terminate( model );
      return (-1);
    }
  if (errCode == 1)  // only the model was read
    {
      if (mpl_read_data( model, datafile ) != 2)  // errors
	{
	  mpl_terminate( model );
	  return (-1);
	}
    }
  else // both the model and data read from modelfile
    {
      if (datafile != 0)
	{
	  cout << "Warning: a data file is indicated but was not "
	       << "used" << endl;
	}
    }

  if (mpl_generate(model, NULL) != 3)  // 3 means it succeeded
  // the NULL indicates where messages should go; NULL means stdout
    {
      mpl_terminate( model );
      return (-1);
    }

  // At this point, parsing was successful, and we just need to transfer 
  // the information into the OsiSolverInterface

  // Don't use these for now, though they are available for the future
  // char *mpl_get_prob_name(MPL *mpl);  
  // char *mpl_get_row_name(MPL *mpl, int i);
  // char *mpl_get_col_name(MPL *mpl, int j);

  int nrows = mpl_get_num_rows(model);
  int ncols = mpl_get_num_cols(model);
  
  double inf = si->getInfinity();

  // Get column bound information
  double * collb = new double[ncols];
  double * colub = new double[ncols];
  int * colkind = new int[ncols];

  for (int i = 0; i < ncols; i++)
    {
      colkind[i] = mpl_get_col_kind( model, i+1);

      if (colkind[i] == MPL_BIN)  // binary variable
	{
	  collb[i] = 0.0;
	  colub[i] = 1.0;
	}
      else  // continuous or general integer variable
	{
	  int type = mpl_get_col_bnds( model, i+1, &collb[i], &colub[i]);
	  switch (type)
	    {
	    case MPL_FR:
	      collb[i] = -inf;
	      colub[i] = inf;
	      break;
	    case MPL_LO:
	      colub[i] = inf;
	      break;
	    case MPL_UP:
	      collb[i] = -inf;
	      break;
	    case MPL_FX:
	      colub[i] = collb[i];
	      break;
	      //case MPL_DB:
	    }
	}
    }

  // Get row information, including the objective row
  
  // First figure out how much space we need for the constraint matrix.
  // This overestimates a bit because it counts the objective, too.
  int ncoeff = 0;
  for (int i = 0; i < nrows; i++)
    ncoeff += mpl_get_mat_row( model, i+1, NULL, NULL );

  // Allocate space for constraint matrix, which will be row-ordered
  int * start = new int[1+nrows];
  int * index = new int[ncoeff];
  double * value = new double[ncoeff];
  int row_count = 0;
  int coeff_count = 0;

  // Allocate space for a temporary buffer for row coeffs
  int len;
  int * index_buf = new int[1+ncols];
  double * value_buf = new double[1+ncols];

  // Allocate space for other information
  double * objcoeff = new double[ncols];
  double * rowlb = new double[nrows];
  double * rowub = new double[nrows];

  for (int i = 0; i < nrows; i++)
    {
      int kind = mpl_get_row_kind( model, i+1 );
      
      if (kind == MPL_MIN || kind == MPL_MAX)  // objective row
	{
	  si->setObjSense( (kind==MPL_MIN)?1:-1 );
	  si->setDblParam(OsiObjOffset, mpl_get_row_c0(model,i+1) );
	  
	  len = mpl_get_mat_row( model, i+1, index_buf, value_buf );
	  // GLPK starts at index 1 for index and value
	  for (int j = 1; j <= len; j++)
	    objcoeff[index_buf[j]-1] = value_buf[j];
	}
      else // constraint row
	{
	  int type = mpl_get_row_bnds( model, i+1, &rowlb[row_count], &rowub[row_count]);
	  switch (type)
	    {
	    case MPL_FR:
	      rowlb[row_count] = -inf;
	      rowub[row_count] = inf;
	      break;
	    case MPL_LO:
	      rowub[row_count] = inf;
	      break;
	    case MPL_UP:
	      rowlb[row_count] = -inf;
	      break;
	    case MPL_FX:
	      rowub[row_count] = rowlb[row_count];
	      break;
	      //case MPL_DB:
	    }

	  len = mpl_get_mat_row( model, i+1, index_buf, value_buf );

	  // Add this to the matrix we're building
	  start[row_count] = coeff_count;
	  for (int j = 1; j <= len; j++)
	    {
	      index[coeff_count] = index_buf[j]-1;
	      value[coeff_count++] = value_buf[j];
	    }
	  row_count++;
	}
    }

  start[row_count] = coeff_count;  // this is the final entry in start

  // create a matrix based on the values we've collected
  CoinPackedMatrix mat;
  mat.copyOf( false, ncols, row_count, coeff_count, value, index, start, 0 );

  si->loadProblem( mat, collb, colub, objcoeff, rowlb, rowub );

  // Go back and make sure we've set integer variables as such
  int ret_code = 1;  // that's for an LP
  for (int i = 0; i < ncols; i++)
    {
      if (colkind[i] == MPL_INT || colkind[i] == MPL_BIN)
	{
	  si->setInteger( i );
	  ret_code = 2;  // this indicates a MIP
	}
    }

  // Free resources used for parsing
  mpl_terminate(model);
  delete [] colkind;
  delete [] collb;
  delete [] colub;
  delete [] rowlb;
  delete [] rowub;
  delete [] objcoeff;
  delete [] start;
  delete [] index;
  delete [] value;
  delete [] index_buf;
  delete [] value_buf;

  return ret_code; // successfully read
}
